package sun.security.rsa;

import java.security.AlgorithmParameters;
import java.security.InvalidParameterException;
import java.security.DigestException;
import java.security.NoSuchAlgorithmException;
import java.security.spec.MGF1ParameterSpec;
import javax.crypto.BadPaddingException;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.nio.ByteBuffer;
import java.security.InvalidAlgorithmParameterException;
import java.security.SignatureException;
import java.util.Arrays;
import java.security.spec.AlgorithmParameterSpec;
import sun.security.jca.JCAUtil;
import java.security.PrivateKey;
import java.security.InvalidKeyException;
import java.security.interfaces.RSAKey;
import java.security.PublicKey;
import java.security.ProviderException;
import java.security.SecureRandom;
import java.security.spec.PSSParameterSpec;
import java.security.interfaces.RSAPublicKey;
import java.security.interfaces.RSAPrivateKey;
import java.security.MessageDigest;
import java.util.Hashtable;
import java.security.SignatureSpi;

public class RSAPSSSignature extends SignatureSpi
{
    private static final boolean DEBUG = false;
    private static final byte[] EIGHT_BYTES_OF_ZEROS;
    private static final Hashtable<String, Integer> DIGEST_LENGTHS;
    private MessageDigest md;
    private boolean digestReset;
    private RSAPrivateKey privKey;
    private RSAPublicKey pubKey;
    private PSSParameterSpec sigParams;
    private SecureRandom random;
    
    private boolean isDigestEqual(String s, String string) {
        if (s == null || string == null) {
            return false;
        }
        if (string.indexOf("-") != -1) {
            return s.equalsIgnoreCase(string);
        }
        if (s.equals("SHA-1")) {
            return string.equalsIgnoreCase("SHA") || string.equalsIgnoreCase("SHA1");
        }
        StringBuilder sb = new StringBuilder(string);
        if (string.regionMatches(true, 0, "SHA", 0, 3)) {
            string = sb.insert(3, "-").toString();
            return s.equalsIgnoreCase(string);
        }
        throw new ProviderException("Unsupported digest algorithm " + string);
    }
    
    public RSAPSSSignature() {
        this.digestReset = true;
        this.privKey = null;
        this.pubKey = null;
        this.sigParams = null;
        this.md = null;
    }
    
    @Override
    protected void engineInitVerify(PublicKey publicKey) throws InvalidKeyException {
        if (!(publicKey instanceof RSAPublicKey)) {
            throw new InvalidKeyException("key must be RSAPublicKey");
        }
        this.pubKey = (RSAPublicKey)this.isValid((RSAKey)publicKey);
        this.privKey = null;
        this.resetDigest();
        return;
    }
    
    @Override
    protected void engineInitSign(PrivateKey privateKey) throws InvalidKeyException {
        this.engineInitSign(privateKey, null);
    }
    
    @Override
    protected void engineInitSign(PrivateKey privateKey, SecureRandom secureRandom) throws InvalidKeyException {
        if (!(privateKey instanceof RSAPrivateKey)) {
            throw new InvalidKeyException("key must be RSAPrivateKey");
        }
        this.privKey = (RSAPrivateKey)this.isValid((RSAKey)privateKey);
        this.pubKey = null;
        this.random = ((secureRandom == null) ? JCAUtil.getSecureRandom() : secureRandom);
        this.resetDigest();
        return;
    }
    
    private static boolean isCompatible(AlgorithmParameterSpec algorithmParameterSpec, PSSParameterSpec pssParameterSpec) {
        if (algorithmParameterSpec == null) {
            return true;
        }
        if (!(algorithmParameterSpec instanceof PSSParameterSpec)) {
            return false;
        }
        if (pssParameterSpec == null) {
            return true;
        }
        PSSParameterSpec pssParameterSpec2 = (PSSParameterSpec)algorithmParameterSpec;
        if (pssParameterSpec2.getSaltLength() > pssParameterSpec.getSaltLength()) {
            return false;
        }
        PSSParameterSpec pssParameterSpec3 = new PSSParameterSpec(pssParameterSpec2.getDigestAlgorithm(), pssParameterSpec2.getMGFAlgorithm(), pssParameterSpec2.getMGFParameters(), pssParameterSpec.getSaltLength(), pssParameterSpec2.getTrailerField());
        PSSParameters pssParameters = new PSSParameters();
        try {
            pssParameters.engineInit(pssParameterSpec3);
            byte[] engineGetEncoded = pssParameters.engineGetEncoded();
            pssParameters.engineInit(pssParameterSpec);
            return Arrays.equals(engineGetEncoded, pssParameters.engineGetEncoded());
        }
        catch (Exception ex) {
            return false;
        }
    }
    
    private RSAKey isValid(RSAKey rsaKey) throws InvalidKeyException {
        try {
            rsaKey.getParams();
            if (!isCompatible(rsaKey.getParams(), this.sigParams)) {
                throw new InvalidKeyException("Key contains incompatible PSS parameter values");
            }
            if (this.sigParams != null) {
                Integer n = (Integer)RSAPSSSignature.DIGEST_LENGTHS.get(this.sigParams.getDigestAlgorithm());
                if (n == null) {
                    throw new ProviderException("Unsupported digest algo: " + this.sigParams.getDigestAlgorithm());
                }
                checkKeyLength(rsaKey, (int)n, this.sigParams.getSaltLength());
            }
            return rsaKey;
        }
        catch (SignatureException cause) {
            throw new InvalidKeyException(cause);
        }
    }
    
    private PSSParameterSpec validateSigParams(AlgorithmParameterSpec algorithmParameterSpec) throws InvalidAlgorithmParameterException {
        if (algorithmParameterSpec == null) {
            throw new InvalidAlgorithmParameterException("Parameters cannot be null");
        }
        if (!(algorithmParameterSpec instanceof PSSParameterSpec)) {
            throw new InvalidAlgorithmParameterException("parameters must be type PSSParameterSpec");
        }
        PSSParameterSpec pssParameterSpec = (PSSParameterSpec)algorithmParameterSpec;
        if (pssParameterSpec == this.sigParams) {
            return pssParameterSpec;
        }
        Object o = (this.privKey == null) ? this.pubKey : this.privKey;
        if (o != null && !isCompatible(o.getParams(), pssParameterSpec)) {
            throw new InvalidAlgorithmParameterException("Signature parameters does not match key parameters");
        }
        if (!pssParameterSpec.getMGFAlgorithm().equalsIgnoreCase("MGF1")) {
            throw new InvalidAlgorithmParameterException("Only supports MGF1");
        }
        if (pssParameterSpec.getTrailerField() != 1) {
            throw new InvalidAlgorithmParameterException("Only supports TrailerFieldBC(1)");
        }
        String digestAlgorithm = pssParameterSpec.getDigestAlgorithm();
        if (o != null) {
            try {
                checkKeyLength(o, (int)RSAPSSSignature.DIGEST_LENGTHS.get(digestAlgorithm), pssParameterSpec.getSaltLength());
            }
            catch (SignatureException cause) {
                throw new InvalidAlgorithmParameterException(cause);
            }
        }
        return pssParameterSpec;
    }
    
    private void ensureInit() throws SignatureException {
        if (((this.privKey == null) ? this.pubKey : this.privKey) == null) {
            throw new SignatureException("Missing key");
        }
        if (this.sigParams == null) {
            throw new SignatureException("Parameters required for RSASSA-PSS signatures");
        }
        return;
    }
    
    private static void checkKeyLength(RSAKey rsaKey, int x, int y) throws SignatureException {
        if (rsaKey != null) {
            int n = getKeyLengthInBits(rsaKey) + 7 >> 3;
            int addExact = Math.addExact(Math.addExact(x, y), 2);
            if (n < addExact) {
                throw new SignatureException("Key is too short, need min " + addExact + " bytes");
            }
        }
    }
    
    private void resetDigest() {
        if (!this.digestReset) {
            this.md.reset();
            this.digestReset = true;
        }
    }
    
    private byte[] getDigestValue() {
        this.digestReset = true;
        return this.md.digest();
    }
    
    @Override
    protected void engineUpdate(byte input) throws SignatureException {
        this.ensureInit();
        this.md.update(input);
        this.digestReset = false;
    }
    
    @Override
    protected void engineUpdate(byte[] input, int offset, int len) throws SignatureException {
        this.ensureInit();
        this.md.update(input, offset, len);
        this.digestReset = false;
    }
    
    @Override
    protected void engineUpdate(ByteBuffer input) {
        try {
            this.ensureInit();
        }
        catch (SignatureException ex) {
            throw new RuntimeException(ex.getMessage());
        }
        this.md.update(input);
        this.digestReset = false;
    }
    
    @Override
    protected byte[] engineSign() throws SignatureException {
        this.ensureInit();
        byte[] digestValue = this.getDigestValue();
        try {
            return RSACore.rsa(this.encodeSignature(digestValue), this.privKey, true);
        }
        catch (GeneralSecurityException cause) {
            throw new SignatureException("Could not sign data", cause);
        }
        catch (IOException cause2) {
            throw new SignatureException("Could not encode data", cause2);
        }
    }
    
    @Override
    protected boolean engineVerify(byte[] array) throws SignatureException {
        this.ensureInit();
        try {
            if (array.length != RSACore.getByteLength(this.pubKey)) {
                throw new SignatureException("Signature length not correct: got " + array.length + " but was expecting " + RSACore.getByteLength(this.pubKey));
            }
            return this.decodeSignature(this.getDigestValue(), RSACore.rsa(array, this.pubKey));
        }
        catch (BadPaddingException ex) {
            return false;
        }
        catch (IOException cause) {
            throw new SignatureException("Signature encoding error", cause);
        }
        finally {
            this.resetDigest();
        }
    }
    
    private static int getKeyLengthInBits(RSAKey rsaKey) {
        if (rsaKey != null) {
            return rsaKey.getModulus().bitLength();
        }
        return -1;
    }
    
    private byte[] encodeSignature(byte[] input) throws IOException, DigestException {
        AlgorithmParameterSpec mgfParameters = this.sigParams.getMGFParameters();
        String s;
        if (mgfParameters != null) {
            s = ((MGF1ParameterSpec)mgfParameters).getDigestAlgorithm();
        }
        else {
            s = this.md.getAlgorithm();
        }
        try {
            int n = getKeyLengthInBits(this.privKey) - 1;
            int n2 = n + 7 >> 3;
            int digestLength = this.md.getDigestLength();
            int offset = n2 - digestLength - 1;
            int saltLength = this.sigParams.getSaltLength();
            byte[] buf = new byte[n2];
            buf[offset - saltLength - 1] = 1;
            buf[buf.length - 1] = -68;
            if (!this.digestReset) {
                throw new ProviderException("Digest should be reset");
            }
            this.md.update(RSAPSSSignature.EIGHT_BYTES_OF_ZEROS);
            this.digestReset = false;
            this.md.update(input);
            if (saltLength != 0) {
                byte[] array = new byte[saltLength];
                this.random.nextBytes(array);
                this.md.update(array);
                System.arraycopy(array, 0, buf, offset - saltLength, saltLength);
            }
            this.md.digest(buf, offset, digestLength);
            this.digestReset = true;
            new MGF1(s).generateAndXor(buf, offset, digestLength, offset, buf, 0);
            int n3 = (n2 << 3) - n;
            if (n3 != 0) {
                buf[0] &= (byte)(255 >>> n3);
                return buf;
            }
            return buf;
        }
        catch (NoSuchAlgorithmException ex) {
            throw new IOException(ex.toString());
        }
    }
    
    private boolean decodeSignature(byte[] input, byte[] array) throws IOException {
        int length = input.length;
        int saltLength = this.sigParams.getSaltLength();
        int n = getKeyLengthInBits(this.pubKey) - 1;
        int n2 = n + 7 >> 3;
        int i = array.length - n2;
        if (i == 1 && array[0] != 0) {
            return false;
        }
        if (n2 < length + saltLength + 2) {
            return false;
        }
        if (array[i + n2 - 1] != -68) {
            return false;
        }
        int n3 = (n2 << 3) - n;
        if (n3 != 0 && (array[i] & (byte)(255 << 8 - n3)) != 0x0) {
            return false;
        }
        AlgorithmParameterSpec mgfParameters = this.sigParams.getMGFParameters();
        String s;
        if (mgfParameters != null) {
            s = ((MGF1ParameterSpec)mgfParameters).getDigestAlgorithm();
        }
        else {
            s = this.md.getAlgorithm();
        }
        int n4 = n2 - length - 1;
        try {
            new MGF1(s).generateAndXor(array, i + n4, length, n4, array, i);
        }
        catch (NoSuchAlgorithmException ex) {
            throw new IOException(ex.toString());
        }
        if (n3 != 0) {
            array[i] &= (byte)(255 >>> n3);
        }
        while (i < i + (n4 - saltLength - 1)) {
            if (array[i] != 0) {
                return false;
            }
            ++i;
        }
        if (array[i] != 1) {
            return false;
        }
        this.md.update(RSAPSSSignature.EIGHT_BYTES_OF_ZEROS);
        this.digestReset = false;
        this.md.update(input);
        if (saltLength > 0) {
            this.md.update(array, i + (n4 - saltLength), saltLength);
        }
        byte[] digest = this.md.digest();
        this.digestReset = true;
        return MessageDigest.isEqual(digest, Arrays.copyOfRange(array, i + n4, i + n2 - 1));
    }
    
    @Deprecated
    @Override
    protected void engineSetParameter(String s, Object o) throws InvalidParameterException {
        throw new UnsupportedOperationException("setParameter() not supported");
    }
    
    @Override
    protected void engineSetParameter(AlgorithmParameterSpec algorithmParameterSpec) throws InvalidAlgorithmParameterException {
        this.sigParams = this.validateSigParams(algorithmParameterSpec);
        if (!this.digestReset) {
            throw new ProviderException("Cannot set parameters during operations");
        }
        String digestAlgorithm = this.sigParams.getDigestAlgorithm();
        if (this.md != null && this.md.getAlgorithm().equalsIgnoreCase(digestAlgorithm)) {
            return;
        }
        try {
            this.md = MessageDigest.getInstance(digestAlgorithm);
        }
        catch (NoSuchAlgorithmException cause) {
            throw new InvalidAlgorithmParameterException("Unsupported digest algorithm " + digestAlgorithm, cause);
        }
        return;
    }
    
    @Deprecated
    @Override
    protected Object engineGetParameter(String s) throws InvalidParameterException {
        throw new UnsupportedOperationException("getParameter() not supported");
    }
    
    @Override
    protected AlgorithmParameters engineGetParameters() {
        AlgorithmParameters instance = null;
        if (this.sigParams != null) {
            try {
                instance = AlgorithmParameters.getInstance("RSASSA-PSS");
                instance.init(this.sigParams);
            }
            catch (GeneralSecurityException ex) {
                throw new ProviderException(ex.getMessage());
            }
        }
        return instance;
    }
    
    static {
        EIGHT_BYTES_OF_ZEROS = new byte[8];
        (DIGEST_LENGTHS = new Hashtable()).put("SHA-1", Integer.valueOf(20));
        RSAPSSSignature.DIGEST_LENGTHS.put("SHA", Integer.valueOf(20));
        RSAPSSSignature.DIGEST_LENGTHS.put("SHA1", Integer.valueOf(20));
        RSAPSSSignature.DIGEST_LENGTHS.put("SHA-224", Integer.valueOf(28));
        RSAPSSSignature.DIGEST_LENGTHS.put("SHA224", Integer.valueOf(28));
        RSAPSSSignature.DIGEST_LENGTHS.put("SHA-256", Integer.valueOf(32));
        RSAPSSSignature.DIGEST_LENGTHS.put("SHA256", Integer.valueOf(32));
        RSAPSSSignature.DIGEST_LENGTHS.put("SHA-384", Integer.valueOf(48));
        RSAPSSSignature.DIGEST_LENGTHS.put("SHA384", Integer.valueOf(48));
        RSAPSSSignature.DIGEST_LENGTHS.put("SHA-512", Integer.valueOf(64));
        RSAPSSSignature.DIGEST_LENGTHS.put("SHA512", Integer.valueOf(64));
        RSAPSSSignature.DIGEST_LENGTHS.put("SHA-512/224", Integer.valueOf(28));
        RSAPSSSignature.DIGEST_LENGTHS.put("SHA512/224", Integer.valueOf(28));
        RSAPSSSignature.DIGEST_LENGTHS.put("SHA-512/256", Integer.valueOf(32));
        RSAPSSSignature.DIGEST_LENGTHS.put("SHA512/256", Integer.valueOf(32));
    }
}
